import json
import os
import signal
import subprocess
import sys
import typing

import cherrypy
import flask
from flask_babel import Babel
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_login import LoginManager, current_user
from flask_mail import Mail
import orjson
from flask_orjson import OrjsonProvider
from flask_sqlalchemy import SQLAlchemy
from bs4 import BeautifulSoup

login_manager = LoginManager()
login_manager.session_protection = 'basic'

mail = Mail()
db = SQLAlchemy(
    engine_options={'future': True},
    session_options={'future': True}
)


def babel_locale_selector() -> str:
    if hasattr(flask.g, 'override_locale') and flask.g.override_locale in sampledb.logic.locale.get_allowed_language_codes():
        return typing.cast(str, flask.g.override_locale)

    request_locale = sampledb.logic.locale.guess_request_locale()

    if not current_user or not current_user.is_authenticated:
        return request_locale

    auto_lc = current_user.settings['AUTO_LC']
    if auto_lc:
        return request_locale

    stored_locale = current_user.settings['LOCALE']
    if stored_locale in sampledb.logic.locale.get_allowed_language_codes():
        return typing.cast(str, stored_locale)

    return request_locale


def babel_timezone_selector() -> typing.Optional[str]:
    if flask.current_app.config['TIMEZONE']:
        return typing.cast(typing.Optional[str], flask.current_app.config['TIMEZONE'])
    if current_user.is_authenticated:
        if current_user.settings['AUTO_TZ']:
            return typing.cast(typing.Optional[str], flask.request.args.get('timezone', current_user.settings['TIMEZONE']))
        return typing.cast(typing.Optional[str], current_user.settings['TIMEZONE'])
    return flask.request.args.get('timezone', None)


babel = Babel()


import sampledb.dashboard
import sampledb.frontend
import sampledb.api
import sampledb.logic
import sampledb.models
import sampledb.models.migrations
import sampledb.config


def setup_database(app: flask.Flask) -> None:
    with app.app_context():
        db.metadata.create_all(bind=db.engine)
        sampledb.models.Objects.bind = db.engine
        sampledb.models.migrations.run(db)


def setup_admin_account_from_config(app: flask.Flask) -> None:
    with app.app_context():
        if 'ADMIN_INFO' in app.config['internal'] and not sampledb.logic.users.get_users(exclude_hidden=False):
            admin_username, admin_email, admin_password = app.config['internal']['ADMIN_INFO']
            admin_user = sampledb.logic.users.create_user(
                admin_username,
                admin_email,
                sampledb.models.users.UserType.PERSON
            )
            sampledb.logic.authentication.add_other_authentication(
                admin_user.id,
                admin_username,
                admin_password,
                confirmed=True
            )
            sampledb.logic.users.set_user_administrator(admin_user.id, True)


def setup_jinja_environment(app: flask.Flask) -> None:
    with app.app_context():
        is_ldap_configured = sampledb.logic.ldap.is_ldap_configured()
        is_oidc_configured = sampledb.logic.oidc.is_oidc_configured()
        is_oidc_only_auth_method = sampledb.logic.oidc.is_oidc_only_auth_method()

    if app.config['JUPYTERHUB_TEMPLATES_URL']:
        jupyterhub_templates_url = app.config['JUPYTERHUB_TEMPLATES_URL']
    elif app.config['JUPYTERHUB_URL']:
        jupyterhub_templates_url = app.config['JUPYTERHUB_URL'] + '/templates'
    else:
        jupyterhub_templates_url = None

    app.jinja_env.add_extension('jinja2.ext.do')
    app.jinja_env.policies['json.dumps_kwargs'] = {'ensure_ascii': False}

    app.jinja_env.globals.update(
        export_file_formats=sampledb.logic.export.FILE_FORMATS,
        jupyterhub_name=app.config['JUPYTERHUB_NAME'],
        jupyterhub_templates_url=jupyterhub_templates_url,
        signout_form=sampledb.frontend.users_forms.SignoutForm,
        service_name=app.config['SERVICE_NAME'],
        service_description=app.config['SERVICE_DESCRIPTION'],
        service_legal_notice=app.config['SERVICE_LEGAL_NOTICE'],
        service_privacy_policy=app.config['SERVICE_PRIVACY_POLICY'],
        service_accessibility=app.config['SERVICE_ACCESSIBILITY'],
        ldap_name=app.config['LDAP_NAME'],
        is_ldap_configured=is_ldap_configured,
        is_oidc_configured=is_oidc_configured,
        is_oidc_only_auth_method=is_oidc_only_auth_method,
        get_action_types=sampledb.logic.action_types.get_action_types,
        get_translated_text=sampledb.logic.utils.get_translated_text,
        get_topics=sampledb.logic.topics.get_topics,
        BeautifulSoup=BeautifulSoup,
        json=json,
        contact_email=app.config['CONTACT_EMAIL'],
        get_user_settings=lambda: sampledb.logic.settings.get_user_settings(current_user.id),
        TimezoneForm=sampledb.frontend.timezone.TimezoneForm,
        get_user_language=sampledb.logic.languages.get_user_language,
        NotificationType=sampledb.models.NotificationType,
        get_user=sampledb.logic.users.get_user,
    )
    app.jinja_env.filters.update(sampledb.frontend.utils.JinjaFilter.filters)
    app.jinja_env.globals.update(sampledb.frontend.utils.JinjaFunction.functions)


def build_translations(pybabel_path: str) -> None:
    translations_directory = os.path.abspath(os.path.join(os.path.dirname(__file__), 'translations'))
    # merge extracted and manual message catalogs
    for translation_directory in os.listdir(translations_directory):
        translation_directory = os.path.join(translations_directory, translation_directory)
        extracted_message_catalog_path = os.path.join(translation_directory, 'LC_MESSAGES', 'extracted_messages.po')
        manual_message_catalog_path = os.path.join(translation_directory, 'LC_MESSAGES', 'manual_messages.po')
        merged_message_catalog_path = os.path.join(translation_directory, 'LC_MESSAGES', 'messages.po')
        if os.path.isfile(extracted_message_catalog_path) and os.path.isfile(manual_message_catalog_path):
            # manually merge instead of using msgcat to avoid duplicate headers
            with open(extracted_message_catalog_path, 'r', encoding='utf-8') as extracted_message_catalog_file:
                extracted_message_catalog = extracted_message_catalog_file.read()
            with open(manual_message_catalog_path, 'r', encoding='utf-8') as manual_message_catalog_file:
                manual_message_catalog = manual_message_catalog_file.read()
            with open(merged_message_catalog_path, 'w', encoding='utf-8') as merged_message_catalog_file:
                merged_message_catalog_file.write(extracted_message_catalog)
                merged_message_catalog_file.write(manual_message_catalog)
    # compile messages
    subprocess.run([pybabel_path, "compile", "-d", translations_directory], check=True)


def create_app(include_dashboard: bool = True) -> flask.Flask:
    app = flask.Flask(__name__)
    app.wsgi_app = ProxyFix(app.wsgi_app)  # type: ignore

    json_provider = OrjsonProvider(app)
    json_provider.option = orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_NAIVE_UTC | orjson.OPT_NON_STR_KEYS
    app.json = json_provider
    app.config.from_object(sampledb.config)

    internal_config = sampledb.config.check_config(app.config)
    app.config['internal'] = internal_config

    login_manager.init_app(app)
    mail.init_app(app)
    db.init_app(app)
    babel.init_app(app, locale_selector=babel_locale_selector, timezone_selector=babel_timezone_selector)

    app.register_blueprint(sampledb.api.server.api)
    app.register_blueprint(sampledb.api.federation.federation_api)
    app.register_blueprint(sampledb.api.frontend.frontend_api)

    app.register_blueprint(sampledb.frontend.frontend)

    if include_dashboard and app.config['ENABLE_MONITORINGDASHBOARD']:
        sampledb.dashboard.init_app(app)

    login_manager.login_view = 'frontend.sign_in'
    login_manager.anonymous_user = sampledb.logic.users.AnonymousUser

    with app.app_context():
        if sampledb.logic.oidc.is_oidc_configured():
            sampledb.logic.oidc.init_app(app)

            if sampledb.logic.oidc.is_oidc_only_auth_method():
                login_manager.login_message = ''

    def custom_send_static_file(filename: str) -> sampledb.utils.FlaskResponseT:
        response = flask.make_response(
            flask.send_from_directory(app.static_folder, filename)  # type: ignore
        )
        if 'v' in flask.request.args:
            # fingerprinted URLs for static files can be cached as immutable
            response.headers['Cache-Control'] = 'public, max-age=31536000, immutable'
        return response

    app.send_static_file = custom_send_static_file  # type: ignore

    setup_database(app)
    setup_admin_account_from_config(app)
    setup_jinja_environment(app)
    if app.config['BUILD_TRANSLATIONS']:
        build_translations(app.config['PYBABEL_PATH'])

    if app.config['ENABLE_NUMERIC_TAGS'] is None:
        with app.app_context():
            # if ENABLE_NUMERIC_TAGS is not set explicitly, only enable numeric tags if they already exist
            app.config['ENABLE_NUMERIC_TAGS'] = sampledb.logic.utils.do_numeric_tags_exist()

    with app.app_context():
        sampledb.logic.utils.print_deprecation_warnings()

        if app.config['ENABLE_BACKGROUND_TASKS']:
            sampledb.logic.background_tasks.reset_claimed_background_tasks()
            sampledb.logic.background_tasks.start_handler_threads(app)

    def signal_handler(sig: int, _: typing.Any) -> None:
        if sig == signal.SIGTERM:
            if app.config['ENABLE_BACKGROUND_TASKS']:
                sampledb.logic.background_tasks.stop_handler_threads(app)
            sys.exit(0)

    signal.signal(signal.SIGTERM, signal_handler)

    @app.route('/dashboard/telemetry/get_is_telemetry_answered')
    def get_is_telemetry_answered() -> sampledb.utils.FlaskResponseT:
        # disable broken Flask-MonitoringDashboard telemetry dialog
        return flask.jsonify({'is_telemetry_answered': True})

    app.csp_reports = []  # type: ignore[attr-defined]

    @app.route('/csp-violation-report', methods=['POST'])
    def csp_report() -> sampledb.utils.FlaskResponseT:
        if app.config.get('TESTING', True):
            app.csp_reports.append(flask.request.get_json(force=True))  # type: ignore[attr-defined]
        elif app.config.get('ENABLE_CONTENT_SECURITY_POLICY', True):
            cherrypy.log("CSP violation report: " + json.dumps(flask.request.get_json(force=True), indent=2))
        return ''

    @app.after_request
    def set_csp_header(response: flask.Response) -> flask.Response:
        default_sources = [
            "'self'",
        ]
        image_sources = [
            "'self'",
            "https:",
            "http:",
            "blob:",
            "data:",
        ]
        script_sources = [
            "'self'",
            "'strict-dynamic'",
            f"'nonce-{sampledb.utils.generate_content_security_policy_nonce()}'",
        ]
        style_sources = [
            "'self'",
            "'unsafe-inline'",
        ]
        connect_sources = [
            "'self'",
        ]
        base_uri = 'none'
        if flask.request.blueprint == 'dashboard':
            # strict-dynamic would block 'unsafe-eval' and sources other than hashes and nonces
            script_sources.remove("'strict-dynamic'")
            script_sources += [
                "'unsafe-eval'",
                "https://cdnjs.cloudflare.com/ajax/libs/d3/",
                "https://ajax.googleapis.com/ajax/libs/angularjs/",
                "https://cdnjs.cloudflare.com/ajax/libs/angular.js/",
                "https://unpkg.com/sunburst-chart",
            ]
            connect_sources += [
                "https://pypi.org/pypi/Flask-MonitoringDashboard/json",
            ]
            base_uri = 'self'
        if flask.request.endpoint == 'frontend.federation':
            connect_sources += [
                '*'
            ]
        elif flask.request.endpoint == 'frontend.federated_login':
            script_sources = [
                "'unsafe-inline'"
            ]
        content_security_policy = f"base-uri '{base_uri}'; default-src {' '.join(default_sources)}; img-src {' '.join(image_sources)}; script-src {' '.join(script_sources)}; style-src {' '.join(style_sources)}; connect-src {' '.join(connect_sources)}; object-src 'none'; report-uri /csp-violation-report"
        if app.config.get('TESTING', True):
            response.headers["Content-Security-Policy-Report-Only"] = content_security_policy
        elif app.config.get('ENABLE_CONTENT_SECURITY_POLICY', True):
            response.headers["Content-Security-Policy"] = content_security_policy
        return response

    return app



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import relationship, Query, Mapped

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .authentication import Authentication


@enum.unique
class HTTPMethod(enum.Enum):
    GET = 0
    POST = 1
    PUT = 2
    PATCH = 3
    DELETE = 4
    HEAD = 5
    OPTIONS = 6
    OTHER = 7

    @staticmethod
    def from_name(
            name: str
    ) -> 'HTTPMethod':
        for member in HTTPMethod:
            if member.name.lower() == name.lower():
                return member
        return HTTPMethod.OTHER


class APILogEntry(Model):
    __tablename__ = 'api_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    api_token_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('authentications.id'), nullable=False)
    method: Mapped[HTTPMethod] = db.Column(db.Enum(HTTPMethod), nullable=False)
    route: Mapped[str] = db.Column(db.String, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    api_token: Mapped['Authentication'] = relationship('Authentication', backref=db.backref("api_log_entries", cascade="all,delete"))

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["APILogEntry"]]

    def __init__(
            self,
            api_token_id: int,
            method: HTTPMethod,
            route: str,
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            api_token_id=api_token_id,
            method=method,
            route=route,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, api_token_id={self.api_token_id}, method={self.method}, route={self.route}, utc_datetime={self.utc_datetime})>'



# coding: utf-8
"""
Add confirmed column to object location assignments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('object_location_assignments', 'confirmed'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ADD confirmed BOOLEAN
    """))
    db.session.execute(db.text("""
        UPDATE object_location_assignments
        SET confirmed = FALSE
        WHERE confirmed IS NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN confirmed SET NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN confirmed SET DEFAULT FALSE
    """))
    return True



# coding: utf-8
"""
RESTful API for the SampleDB frontend
"""

from flask import Blueprint

frontend_api = Blueprint('frontend_api', __name__)

from . import favorites

frontend_api.add_url_rule('/api/frontend/favorite_actions/<int:action_id>', endpoint='favorite_actions', view_func=favorites.FavoriteAction.as_view('favorite_action'))
frontend_api.add_url_rule('/api/frontend/favorite_instruments/<int:instrument_id>', endpoint='favorite_instruments', view_func=favorites.FavoriteInstrument.as_view('favorite_instrument'))



# coding: utf-8
"""
Replace the public_actions table with the all_user_action_permissions table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    table_exists = db.session.execute(db.text("""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = 'public_actions'
    """)).fetchall()
    if not table_exists:
        return False

    # Perform migration
    public_actions = db.session.execute(db.text("""
        SELECT action_id
        FROM public_actions
    """)).fetchall()
    all_user_action_permissions_result = db.session.execute(db.text("""
        SELECT action_id, permissions
        FROM all_user_action_permissions
    """)).fetchall()
    all_user_action_permissions = {
        action_id: permissions
        for action_id, permissions in all_user_action_permissions_result
    }
    for action_id, in public_actions:
        if action_id not in all_user_action_permissions:
            db.session.execute(db.text("""
                INSERT INTO all_user_action_permissions
                (action_id, permissions)
                VALUES (:action_id, 'READ')
            """), {'action_id': action_id})
    db.session.execute(db.text("""
        DROP TABLE public_actions
    """))
    return True



# coding: utf-8
"""
RESTful API for SampleDB
"""

import json
import typing

import flask

from ..utils import Resource, ResponseData
from ...utils import text_to_bool
from ...api.server.authentication import multi_auth, object_permissions_required
from ...logic.action_types import check_action_type_exists
from ...logic.actions import get_action, check_action_exists
from ...logic.action_permissions import get_user_action_permissions
from ...logic.object_search import generate_filter_func, wrap_filter_func
from ...logic.objects import get_object, update_object, create_object
from ...logic.object_permissions import get_objects_with_permissions
from ...logic.object_relationships import get_referencing_object_ids, get_related_object_ids
from ...logic.schemas.data_diffs import apply_diff, calculate_diff
from ...logic import errors, users
from ... import models
from ...models import Permissions

from .users import user_to_json
from .actions import action_to_json

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class ObjectVersion(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, version_id: int) -> ResponseData:
        try:
            object = get_object(object_id=object_id, version_id=version_id)
        except errors.ObjectVersionDoesNotExistError:
            return {
                "message": f"version {version_id} of object {object_id} does not exist"
            }, 404
        object_version_json: typing.Dict[str, typing.Any] = {
            'object_id': object.object_id,
            'version_id': object.version_id,
            'action_id': object.action_id,
            'user_id': object.user_id,
            'utc_datetime': object.utc_datetime.strftime("%Y-%m-%d %H:%M:%S") if object.utc_datetime else None,
            'schema': object.schema,
            'data': object.data,
            'fed_object_id': object.fed_object_id,
            'fed_version_id': object.fed_version_id,
            'component_id': object.component_id
        }
        embed_action = bool(flask.request.args.get('embed_action'))
        if embed_action:
            object_version_json['action'] = None
            if object.action_id is not None:
                try:
                    if Permissions.READ in get_user_action_permissions(action_id=object.action_id, user_id=flask.g.user.id):
                        action = get_action(
                            action_id=object.action_id
                        )
                        object_version_json['action'] = action_to_json(action)
                except errors.ActionDoesNotExistError:
                    pass
        embed_user = bool(flask.request.args.get('embed_user'))
        if embed_user:
            object_version_json['user'] = None
            if object.user_id is not None:
                try:
                    user = users.get_user(object.user_id)
                    object_version_json['user'] = user_to_json(user)
                except errors.UserDoesNotExistError:
                    pass
        include_diff = text_to_bool(flask.request.args.get('include_diff', ''))
        if include_diff and object.version_id > 0:
            previous_object_version = get_object(object_id=object.object_id, version_id=object.version_id - 1)
            data_diff = calculate_diff(previous_object_version.data, object.data)
            object_version_json['data_diff'] = data_diff
        return object_version_json


class ObjectVersions(Resource):
    @object_permissions_required(Permissions.WRITE)
    def post(self, object_id: int) -> ResponseData:
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        for key in request_json:
            if key not in {'object_id', 'fed_object_id', 'fed_version_id', 'component_id', 'version_id', 'action_id', 'schema', 'data', 'data_diff'}:
                return {
                    "message": f"invalid key '{key}'"
                }, 400
        object = get_object(object_id=object_id)
        if object.action_id is None:
            return {
                "message": "editing this object is not supported"
            }, 400
        action = get_action(action_id=object.action_id)
        if 'object_id' in request_json:
            if request_json['object_id'] != object.object_id:
                return {
                    "message": f"object_id must be {object.object_id}"
                }, 400
        if 'component_id' in request_json:
            if request_json['component_id'] != object.component_id:
                return {
                    "message": f"component_id must be {object.component_id}"
                }, 400
        if 'fed_version_id' in request_json:
            if request_json['fed_version_id'] != object.fed_version_id:
                return {
                    "message": f"fed_version_id must be {object.fed_version_id}"
                }, 400
        if 'fed_object_id' in request_json:
            if request_json['fed_object_id'] != object.fed_object_id:
                return {
                    "message": f"fed_object_id must be {object.fed_object_id}"
                }, 400
        if 'version_id' in request_json:
            if request_json['version_id'] != object.version_id + 1:
                return {
                    "message": f"version_id must be {object.version_id + 1}"
                }, 400
        if 'action_id' in request_json:
            if request_json['action_id'] != object.action_id:
                return {
                    "message": f"action_id must be {object.action_id}"
                }, 400
        if 'schema' in request_json:
            if request_json['schema'] != object.schema and request_json['schema'] != action.schema:
                return {
                    "message": f"schema must be either:\n{json.dumps(object.schema, indent=4)}\nor:\n{json.dumps(action.schema, indent=4)}"
                }, 400
            schema = request_json['schema']
        else:
            schema = object.schema

        data = request_json.get('data')
        data_diff = request_json.get('data_diff')
        try:
            if data_diff is not None:
                if object.data is None or object.schema is None:
                    return {
                        "message": "previous object version must contain data and schema"
                    }, 400
                data_from_diff = apply_diff(object.data, data_diff, object.schema)
            else:
                data_from_diff = None
            if data is not None and data_diff is not None:
                if data_from_diff != data:
                    return {
                        "message": "data and data_diff are conflicting"
                    }, 400
            if data is None:
                if data_from_diff is None:
                    return {
                        "message": "data or data_diff must be set"
                    }, 400
                else:
                    data = data_from_diff

            update_object(
                object_id=object.id,
                data=data,
                user_id=flask.g.user.id,
                schema=schema
            )

        except errors.ValidationError as e:
            messages = e.message.splitlines()
            return {
                "message": "validation failed:\n - " + "\n - ".join(messages),
                "error_paths": e.paths
            }, 400
        except errors.DiffMismatchError:
            return {
                "message": "failed to apply diff"
            }, 400
        except Exception:
            return {
                "message": "failed to update object"
            }, 400

        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id + 1,
            _external=True
        )

        return flask.redirect(object_version_url, code=201)


class Object(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int) -> ResponseData:
        object = get_object(object_id=object_id)
        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id,
            _external=True
        )
        return flask.redirect(object_version_url, code=302)


class Objects(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        action_id: typing.Optional[int] = None
        action_id_str = flask.request.args.get('action_id', '')
        if action_id_str:
            try:
                action_id = int(action_id_str)
            except ValueError:
                return {
                    'message': 'Unable to parse action_id'
                }, 400
            try:
                check_action_exists(action_id)
            except errors.ActionDoesNotExistError:
                return {
                    'message': 'No action with the given action_id exists.'
                }, 400

        action_type_id: typing.Optional[int] = None
        action_type_id_str = flask.request.args.get('action_type_id', flask.request.args.get('action_type', None))
        if action_type_id_str is not None:
            try:
                action_type_id = int(action_type_id_str)
            except ValueError:
                # ensure old links still function
                action_type_id = {
                    'sample': models.ActionType.SAMPLE_CREATION,
                    'measurement': models.ActionType.MEASUREMENT,
                    'simulation': models.ActionType.SIMULATION
                }.get(action_type_id_str, None)
            else:
                try:
                    check_action_type_exists(action_type_id)
                except errors.ActionTypeDoesNotExistError:
                    action_type_id = None
            if action_type_id is None:
                return {
                    'message': 'No matching action type exists.'
                }, 400

        project_id = None

        limit: typing.Optional[int] = None
        limit_str = flask.request.args.get('limit')
        if limit_str is not None:
            try:
                limit = int(limit_str)
            except ValueError:
                pass
        if limit is not None and not 1 <= limit < 1e15:
            limit = None

        offset: typing.Optional[int] = None
        offset_str = flask.request.args.get('offset')
        if offset_str is not None:
            try:
                offset = int(offset_str)
            except ValueError:
                pass
        if offset is not None and not 0 <= offset < 1e15:
            offset = None

        name_only = text_to_bool(flask.request.args.get('name_only', ''))
        query_string = flask.request.args.get('q', '')
        if query_string:
            try:
                unwrapped_filter_func, _search_tree, _use_advanced_search = generate_filter_func(query_string, text_to_bool(flask.request.args.get('use_advanced_search', 'true')), use_permissions_filter_for_referenced_objects=not flask.g.user.has_admin_permissions)
            except Exception:
                # TODO: ensure that advanced search does not cause exceptions
                def unwrapped_filter_func(data: typing.Any, search_notes: typing.List[typing.Tuple[str, str, int, typing.Optional[int]]]) -> typing.Any:
                    """ Return all objects"""
                    search_notes.append(('error', "Unable to parse search expression", 0, len(query_string)))
                    return False
            filter_func, search_notes = wrap_filter_func(unwrapped_filter_func)
        else:
            search_notes = []

            def filter_func(data: typing.Any) -> typing.Any:
                return True
        try:
            objects = get_objects_with_permissions(
                user_id=flask.g.user.id,
                permissions=Permissions.READ,
                filter_func=filter_func,
                action_ids=[action_id] if action_id is not None else None,
                action_type_ids=[action_type_id] if action_type_id is not None else None,
                project_id=project_id,
                limit=limit,
                offset=offset,
                name_only=name_only
            )
        except Exception as e:
            search_notes.append(('error', f"Error during search: {e}", 0, 0))
            objects = []
        need_object_references = flask.request.args.get('get_referencing_objects', False)
        if any(search_note[0] == 'error' for search_note in search_notes):
            return {
                'message': '\n'.join(
                    'Error: ' + search_note[1]
                    for search_note in search_notes
                    if search_note[0] == 'error'
                )
            }, 400
        elif need_object_references:
            object_references = get_referencing_object_ids({object.id for object in objects})
            ret = []
            for object in objects:
                references = object_references[object.object_id]
                referenced: typing.List[typing.Dict[str, typing.Any]] = []
                for ref in references:
                    if ref is not None:
                        reference_dict: typing.Dict[str, typing.Any] = {'object_id': ref.object_id}
                        if ref.component_uuid is not None:
                            reference_dict['component_uuid'] = ref.component_uuid
                        if ref.eln_source_url is not None:
                            reference_dict['eln_source_url'] = ref.eln_source_url
                        if ref.eln_object_url is not None:
                            reference_dict['eln_object_id'] = ref.eln_object_url
                        referenced.append(reference_dict)
                ret.append(
                    {
                        'object_id': object.object_id,
                        'version_id': object.version_id,
                        'action_id': object.action_id,
                        'user_id': object.user_id,
                        'utc_datetime': object.utc_datetime.strftime("%Y-%m-%d %H:%M:%S") if object.utc_datetime is not None else None,
                        'schema': object.schema,
                        'data': object.data,
                        'fed_object_id': object.fed_object_id,
                        'fed_version_id': object.fed_version_id,
                        'component_id': object.component_id,
                        'referencing_objects': referenced
                    }
                )
            return ret, 200
        else:
            return [
                {
                    'object_id': object.object_id,
                    'version_id': object.version_id,
                    'action_id': object.action_id,
                    'user_id': object.user_id,
                    'utc_datetime': object.utc_datetime.strftime("%Y-%m-%d %H:%M:%S") if object.utc_datetime is not None else None,
                    'schema': object.schema,
                    'data': object.data,
                    'fed_object_id': object.fed_object_id,
                    'fed_version_id': object.fed_version_id,
                    'component_id': object.component_id
                }
                for object in objects
            ], 200

    @multi_auth.login_required
    def post(self) -> ResponseData:
        if flask.g.user.is_readonly:
            return {
                'message': 'user has been marked as read only'
            }, 400
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        for key in request_json:
            if key not in {'object_id', 'fed_object_id', 'fed_version_id', 'component_id', 'version_id', 'action_id', 'schema', 'data'}:
                return {
                    "message": f"invalid key '{key}'"
                }, 400
        if 'object_id' in request_json:
            return {
                "message": "object_id cannot be set"
            }, 400
        if 'fed_object_id' in request_json:
            if request_json['fed_object_id'] is not None:
                return {
                    "message": "fed_object_id must be null"
                }, 400
        if 'fed_version_id' in request_json:
            if request_json['fed_version_id'] is not None:
                return {
                    "message": "fed_version_id must be null"
                }, 400
        if 'component_id' in request_json:
            if request_json['component_id'] is not None:
                return {
                    "message": "component_id must be null"
                }, 400
        if 'version_id' in request_json:
            if request_json['version_id'] != 0:
                return {
                    "message": "version_id must be 0"
                }, 400
        if 'action_id' in request_json:
            if not isinstance(request_json['action_id'], int):
                return {
                    "message": "action_id must be an integer"
                }, 400
            action_id = request_json['action_id']
            try:
                action = get_action(action_id=action_id)
            except errors.ActionDoesNotExistError:
                return {
                    "message": f"action {action_id} does not exist"
                }, 400
            if action.type is None or action.type.disable_create_objects or action.disable_create_objects or (action.admin_only and not flask.g.user.is_admin):
                return {
                    "message": f"creating objects with action {action_id} is disabled"
                }, 400
        else:
            return {
                "message": "action_id must be set"
            }, 400
        if 'schema' in request_json:
            if request_json['schema'] != action.schema:
                return {
                    "message": "schema must be:\n" + json.dumps(action.schema, indent=4)
                }, 400
        schema = action.schema
        if 'data' not in request_json:
            return {
                "message": "data must be set"
            }, 400
        data = request_json['data']
        try:
            object = create_object(
                action_id=action_id,
                data=data,
                user_id=flask.g.user.id,
                schema=schema
            )
        except errors.ValidationError as e:
            messages = e.message.splitlines()
            return {
                "message": "validation failed:\n - " + "\n - ".join(messages),
                "error_paths": e.paths
            }, 400
        except Exception:
            return {
                "message": "failed to create object"
            }, 400
        object_version_url = flask.url_for(
            'api.object_version',
            object_id=object.object_id,
            version_id=object.version_id,
            _external=True
        )
        return flask.redirect(object_version_url, code=201)


class RelatedObjects(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int) -> ResponseData:
        referencing_object_ids, referenced_object_ids = get_related_object_ids(
            object_id=object_id,
            include_referencing_objects=True,
            include_referenced_objects=True,
            user_id=flask.g.user.id
        )
        return {
            "referencing_objects": [
                {
                    "object_id": object_id,
                    "component_uuid": component_uuid
                }
                for object_id, component_uuid in referencing_object_ids
            ],
            "referenced_objects": [
                {
                    "object_id": object_id,
                    "component_uuid": component_uuid
                }
                for object_id, component_uuid in referenced_object_ids
            ]
        }



# coding: utf-8
"""
Script for updating the users responsible for an instrument in SampleDB.

Usage: sampledb update_instrument_responsible_users <instrument_id> <instrument_reponsible_user_ids...>
"""

import sys
import typing

from .. import create_app
from ..logic.instruments import get_instrument, add_instrument_responsible_user, remove_instrument_responsible_user
from ..logic.users import check_user_exists
from ..logic.errors import UserDoesNotExistError, InstrumentDoesNotExistError


def main(arguments: typing.List[str]) -> None:
    if len(arguments) == 0:
        print(__doc__)
        sys.exit(1)
    try:
        instrument_id = int(arguments[0])
    except ValueError:
        print("Error: instrument_id must be an integer", file=sys.stderr)
        sys.exit(1)
    instrument_responsible_user_ids = []
    for user_id_str in arguments[1:]:
        try:
            user_id = int(user_id_str)
        except ValueError:
            print("Error: instrument_reponsible_user_ids must be integer", file=sys.stderr)
            sys.exit(1)
        instrument_responsible_user_ids.append(user_id)
    app = create_app()
    with app.app_context():
        if app.config['DISABLE_INSTRUMENTS']:
            print('Error: instruments are disabled', file=sys.stderr)
            sys.exit(1)
        try:
            instrument = get_instrument(instrument_id)
        except InstrumentDoesNotExistError:
            print('Error: no instrument with this id exists', file=sys.stderr)
            sys.exit(1)
        for user_id in instrument_responsible_user_ids:
            try:
                check_user_exists(user_id)
            except UserDoesNotExistError:
                print(f'Error: no user with the id #{user_id} exists', file=sys.stderr)
                sys.exit(1)
        previous_instrument_responsible_user_ids = [user.id for user in instrument.responsible_users]
        for user_id in instrument_responsible_user_ids:
            if user_id not in previous_instrument_responsible_user_ids:
                add_instrument_responsible_user(instrument.id, user_id)
        for user_id in previous_instrument_responsible_user_ids:
            if user_id not in instrument_responsible_user_ids:
                remove_instrument_responsible_user(instrument.id, user_id)
        print("Success: the instrument responsible users have been updated in SampleDB")



import typing

import flask
from flask_wtf import FlaskForm
from wtforms import SelectField, IntegerField, FieldList, FormField
from wtforms.validators import InputRequired

from ..logic import errors
from ..logic.permissions import ResourcePermissions
from ..models import Permissions


class UserPermissionsForm(FlaskForm):
    user_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class GroupPermissionsForm(FlaskForm):
    group_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class ProjectPermissionsForm(FlaskForm):
    project_id = IntegerField(
        validators=[InputRequired()]
    )
    permissions = SelectField(
        choices=[(p.name.lower(), p.name.lower()) for p in (Permissions.NONE, Permissions.READ, Permissions.WRITE, Permissions.GRANT)],
        validators=[InputRequired()]
    )


class PermissionsForm(FlaskForm):
    all_user_permissions = SelectField(
        choices=[(p.name.lower(), p) for p in (Permissions.NONE, Permissions.READ)],
        validators=[InputRequired()]
    )
    anonymous_user_permissions = SelectField(
        choices=[(p.name.lower(), p) for p in (Permissions.NONE, Permissions.READ)],
        validators=[InputRequired()]
    )
    user_permissions = FieldList(FormField(UserPermissionsForm), min_entries=0)
    group_permissions = FieldList(FormField(GroupPermissionsForm), min_entries=0)
    project_permissions = FieldList(FormField(ProjectPermissionsForm), min_entries=0)


def handle_permission_forms(
        resource_permissions: ResourcePermissions,
        resource_id: int,
        add_user_permissions_form: UserPermissionsForm,
        add_group_permissions_form: GroupPermissionsForm,
        add_project_permissions_form: ProjectPermissionsForm,
        permissions_form: PermissionsForm,
        existing_user_permissions: typing.Optional[typing.Dict[int, Permissions]] = None,
        existing_group_permissions: typing.Optional[typing.Dict[int, Permissions]] = None,
        existing_project_permissions: typing.Optional[typing.Dict[int, Permissions]] = None
) -> bool:
    if 'add_user_permissions' in flask.request.form and add_user_permissions_form.validate_on_submit():
        user_id = add_user_permissions_form.user_id.data
        permissions = Permissions.from_name(add_user_permissions_form.permissions.data)
        if existing_user_permissions is None:
            existing_user_permissions = resource_permissions.get_permissions_for_users(resource_id)
        if permissions != Permissions.NONE and user_id not in existing_user_permissions:
            try:
                resource_permissions.set_permissions_for_user(
                    resource_id=resource_id,
                    user_id=user_id,
                    permissions=permissions
                )
            except errors.UserDoesNotExistError:
                return False
            else:
                return True
    elif 'add_group_permissions' in flask.request.form and add_group_permissions_form.validate_on_submit():
        group_id = add_group_permissions_form.group_id.data
        permissions = Permissions.from_name(add_group_permissions_form.permissions.data)
        if existing_group_permissions is None:
            existing_group_permissions = resource_permissions.get_permissions_for_groups(resource_id)
        if permissions != Permissions.NONE and group_id not in existing_group_permissions:
            try:
                resource_permissions.set_permissions_for_group(
                    resource_id=resource_id,
                    group_id=group_id,
                    permissions=permissions
                )
            except errors.GroupDoesNotExistError:
                return False
            else:
                return True
    elif 'add_project_permissions' in flask.request.form and add_project_permissions_form.validate_on_submit():
        project_id = add_project_permissions_form.project_id.data
        permissions = Permissions.from_name(add_project_permissions_form.permissions.data)
        if existing_project_permissions is None:
            existing_project_permissions = resource_permissions.get_permissions_for_projects(resource_id)
        if permissions != Permissions.NONE and project_id not in existing_project_permissions:
            try:
                resource_permissions.set_permissions_for_project(
                    resource_id=resource_id,
                    project_id=project_id,
                    permissions=permissions
                )
            except errors.ProjectDoesNotExistError:
                return False
            else:
                return True
    elif 'edit_permissions' in flask.request.form and permissions_form.validate_on_submit():
        permissions = Permissions.from_name(permissions_form.all_user_permissions.data)
        resource_permissions.set_permissions_for_all_users(resource_id, permissions)
        permissions = Permissions.from_name(permissions_form.anonymous_user_permissions.data)
        resource_permissions.set_permissions_for_anonymous_users(resource_id, permissions)
        for user_permissions_data in permissions_form.user_permissions.data:
            user_id = user_permissions_data['user_id']
            permissions = Permissions.from_name(user_permissions_data['permissions'])
            try:
                resource_permissions.set_permissions_for_user(
                    resource_id=resource_id,
                    user_id=user_id,
                    permissions=permissions
                )
            except errors.UserDoesNotExistError:
                continue
        for group_permissions_data in permissions_form.group_permissions.data:
            group_id = group_permissions_data['group_id']
            permissions = Permissions.from_name(group_permissions_data['permissions'])
            try:
                resource_permissions.set_permissions_for_group(
                    resource_id=resource_id,
                    group_id=group_id,
                    permissions=permissions
                )
            except errors.GroupDoesNotExistError:
                continue
        for project_permissions_data in permissions_form.project_permissions.data:
            project_id = project_permissions_data['project_id']
            permissions = Permissions.from_name(project_permissions_data['permissions'])
            try:
                resource_permissions.set_permissions_for_project(
                    resource_id=resource_id,
                    project_id=project_id,
                    permissions=permissions
                )
            except errors.ProjectDoesNotExistError:
                continue
        return True
    return False


def set_up_permissions_forms(
        *,
        resource_permissions: ResourcePermissions,
        resource_id: int,
        existing_all_user_permissions: typing.Optional[Permissions] = None,
        existing_anonymous_user_permissions: typing.Optional[Permissions] = None,
        existing_user_permissions: typing.Optional[typing.Dict[int, Permissions]] = None,
        existing_group_permissions: typing.Optional[typing.Dict[int, Permissions]] = None,
        existing_project_permissions: typing.Optional[typing.Dict[int, Permissions]] = None
) -> typing.Tuple[UserPermissionsForm, GroupPermissionsForm, ProjectPermissionsForm, PermissionsForm]:
    if existing_all_user_permissions is None:
        existing_all_user_permissions = resource_permissions.get_permissions_for_all_users(resource_id)
    if existing_anonymous_user_permissions is None:
        existing_anonymous_user_permissions = resource_permissions.get_permissions_for_anonymous_users(resource_id)
    if existing_user_permissions is None:
        existing_user_permissions = resource_permissions.get_permissions_for_users(resource_id)
    if existing_group_permissions is None:
        existing_group_permissions = resource_permissions.get_permissions_for_groups(resource_id)
    if existing_project_permissions is None:
        existing_project_permissions = resource_permissions.get_permissions_for_projects(resource_id)

    all_user_permissions_form_data = existing_all_user_permissions.name.lower()

    anonymous_user_permissions_form_data = existing_anonymous_user_permissions.name.lower()

    user_permission_form_data = []
    for user_id, permissions in sorted(existing_user_permissions.items()):
        if user_id is None:
            continue
        user_permission_form_data.append({'user_id': user_id, 'permissions': permissions.name.lower()})

    group_permission_form_data = []
    for group_id, permissions in sorted(existing_group_permissions.items()):
        if group_id is None:
            continue
        group_permission_form_data.append({'group_id': group_id, 'permissions': permissions.name.lower()})

    project_permission_form_data = []
    for project_id, permissions in sorted(existing_project_permissions.items()):
        if project_id is None:
            continue
        project_permission_form_data.append({'project_id': project_id, 'permissions': permissions.name.lower()})

    permissions_form = PermissionsForm(
        all_user_permissions=all_user_permissions_form_data,
        anonymous_user_permissions=anonymous_user_permissions_form_data,
        user_permissions=user_permission_form_data,
        group_permissions=group_permission_form_data,
        project_permissions=project_permission_form_data
    )
    add_user_permissions_form = UserPermissionsForm()
    add_group_permissions_form = GroupPermissionsForm()
    add_project_permissions_form = ProjectPermissionsForm()
    return (
        add_user_permissions_form,
        add_group_permissions_form,
        add_project_permissions_form,
        permissions_form
    )



# coding: utf-8
"""
RESTful API for SampleDB
"""

import flask

from ..utils import Resource, ResponseData
from ...api.server.authentication import multi_auth
from ...logic.object_log import get_object_log_entries_by_user, object_log_entry_to_json


class ObjectLogEntries(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        after_id = 0
        after_id_str = flask.request.args.get('after_id')
        if after_id_str is not None:
            try:
                after_id = max(int(after_id_str), 0)
            except ValueError:
                pass

        log_entries = get_object_log_entries_by_user(flask.g.user.id, after_id)
        return [object_log_entry_to_json(log_entry) for log_entry in log_entries]



# coding: utf-8
"""

"""

import dataclasses
import datetime
import functools
import typing

import sqlalchemy as db
from sqlalchemy.dialects import postgresql

if typing.TYPE_CHECKING:
    from ..logic.components import Component
    from ..logic.eln_import import ELNImport


F = typing.TypeVar('F', bound=typing.Callable[..., typing.Any])


def _use_transaction(func: F) -> F:
    @functools.wraps(func)
    def wrapped_func(
            self: 'VersionedJSONSerializableObjectTables',
            *args: typing.Any,
            connection: typing.Optional[db.engine.Connection] = None,
            **kwargs: typing.Any
    ) -> typing.Any:
        if connection is not None:
            return func(self, *args, connection=connection, **kwargs)
        assert self.bind is not None
        with self.bind.begin() as connection:
            return func(self, *args, connection=connection, **kwargs)
    return typing.cast(F, wrapped_func)


class DataValidator(typing.Protocol):
    def __call__(
        self,
        instance: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None,
        allow_disabled_languages: bool = False,
        strict: bool = False
    ) -> None:
        ...


class SchemaValidator(typing.Protocol):
    def __call__(
        self,
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None,
        *,
        parent_conditions: typing.Optional[typing.List[typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]]] = None,
        invalid_template_action_ids: typing.Sequence[int] = (),
        strict: bool = False
    ) -> None:
        ...


@dataclasses.dataclass
class Object:
    object_id: int
    version_id: int
    action_id: typing.Optional[int]
    data: typing.Optional[typing.Dict[str, typing.Any]]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    user_id: typing.Optional[int]
    utc_datetime: typing.Optional[datetime.datetime]
    fed_object_id: typing.Optional[int]
    fed_version_id: typing.Optional[int]
    component_id: typing.Optional[int]
    eln_import_id: typing.Optional[int]
    eln_object_id: typing.Optional[str]
    _component_cache: typing.List[typing.Optional['Component']] = dataclasses.field(default_factory=lambda: [None], repr=False, kw_only=True)
    _eln_import_cache: typing.List[typing.Optional['ELNImport']] = dataclasses.field(default_factory=lambda: [None], repr=False, kw_only=True)

    @property
    def id(self) -> int:
        return self.object_id

    @property
    def fed_id(self) -> typing.Optional[int]:
        return self.fed_object_id

    @property
    def name(self) -> typing.Union[str, typing.Dict[str, str]]:
        if self.data is not None and self.data.get('name') and isinstance(self.data['name'], dict) and self.data['name'].get('text'):
            return typing.cast(typing.Union[str, typing.Dict[str, str]], self.data['name']['text'])
        else:
            return ''

    @property
    def component(self) -> typing.Optional['Component']:
        if self.component_id is not None and self._component_cache[0] is None:
            from ..logic.components import get_component
            self._component_cache[0] = get_component(self.component_id)
        return self._component_cache[0]

    @property
    def eln_import(self) -> typing.Optional['ELNImport']:
        if self.eln_import_id is not None and self._eln_import_cache[0] is None:
            from ..logic.eln_import import get_eln_import
            self._eln_import_cache[0] = get_eln_import(self.eln_import_id)
        return self._eln_import_cache[0]


class VersionedJSONSerializableObjectTables:
    """
    A class for storing JSON-serializable objects without deletes and with versioned updates.

    The class uses two tables, one for the current objects and one for previous versions of these objects. Each
    object consists of an object ID, a version ID, the actual data, a JSON schema for the data and information about
    this current version (the ID of this version's author and a datetime of when the object was created or updated).

    These two tables will be created when the instance of this class is created, if bind is provided. Otherwise you
    can use the metadata attribute to access the SQLAlchemy MetaData object associated with these tables. You can then
    call metadata.create_all(bind) to create the tables.

    You should **not** interact with the tables yourself. Instead, use the functions provided by this class.

    For information on JSON schemas, see http://json-schema.org/.
    """

    def __init__(
            self,
            table_name_prefix: str,
            bind: typing.Optional[db.engine.Engine] = None,
            user_id_column: typing.Optional[typing.Any] = None,
            action_id_column: typing.Optional[typing.Any] = None,
            component_id_column: typing.Optional[typing.Any] = None,
            eln_import_id_column: typing.Optional[typing.Any] = None,
            action_schema_column: typing.Optional[typing.Any] = None,
            metadata: typing.Optional[db.MetaData] = None,
            data_validator: typing.Optional[DataValidator] = None,
            schema_validator: typing.Optional[SchemaValidator] = None
    ) -> None:
        """
        Creates new instance for storing versioned, JSON-serializable objects using three tables.

        :param table_name_prefix: the prefix used for naming the two used tables
        :param bind: the SQLAlchemy engine used for creating the tables and for future connections
        :param user_id_column: a SQLAlchemy column object for use as foreign key for the user ID (optional)
        :param action_id_column: a SQLAlchemy column object for use as foreign key for the action ID (optional)
        :param metadata: an SQLAlchemy MetaData object used for creating the two tables (optional)
        :param data_validator: a data validator function (given the data and the schema) (optional)
        :param schema_validator: a schema validator function (given the schema) (optional)
        """
        if metadata is None:
            metadata = db.MetaData()
        self.metadata = metadata
        self._current_table = db.Table(
            table_name_prefix + '_current',
            self.metadata,
            db.Column('object_id', db.Integer, nullable=False, primary_key=True, autoincrement=True),
            db.Column('version_id', db.Integer, nullable=False, default=0),
            db.Column('action_id', db.Integer),
            db.Column('data', postgresql.JSONB),
            db.Column('schema', postgresql.JSONB),
            db.Column('user_id', db.Integer),
            db.Column('utc_datetime', db.TIMESTAMP(timezone=True)),
            db.Column('fed_object_id', db.Integer),
            db.Column('fed_version_id', db.Integer),
            db.Column('component_id', db.Integer),
            db.Column('name_cache', db.JSON),
            db.Column('tags_cache', db.JSON),
            db.Column('eln_import_id', db.Integer, nullable=True),
            db.Column('eln_object_id', db.String, nullable=True),
            db.CheckConstraint(
                '(fed_object_id IS NOT NULL AND fed_version_id IS NOT NULL AND component_id IS NOT NULL) OR (eln_import_id IS NOT NULL AND eln_object_id IS NOT NULL) OR (action_id IS NOT NULL AND data IS NOT NULL AND schema IS NOT NULL AND user_id IS NOT NULL AND utc_datetime IS NOT NULL)',
                name=table_name_prefix + '_current_not_null_check'
            ),
            db.UniqueConstraint('fed_object_id', 'fed_version_id', 'component_id', name=table_name_prefix + '_current_fed_object_id_component_id_key')
        )
        self._previous_table = db.Table(
            table_name_prefix + '_previous',
            self.metadata,
            db.Column('object_id', db.Integer, nullable=False),
            db.Column('version_id', db.Integer, nullable=False),
            db.Column('action_id', db.Integer, nullable=True),
            db.Column('data', postgresql.JSONB, nullable=True),
            db.Column('schema', postgresql.JSONB, nullable=True),
            db.Column('user_id', db.Integer, nullable=True),
            db.Column('utc_datetime', db.TIMESTAMP(timezone=True), nullable=True),
            db.Column('fed_object_id', db.Integer, nullable=True),
            db.Column('fed_version_id', db.Integer, nullable=True),
            db.Column('component_id', db.Integer, nullable=True),
            db.Column('eln_import_id', db.Integer, nullable=True),
            db.Column('eln_object_id', db.String, nullable=True),
            db.PrimaryKeyConstraint('object_id', 'version_id'),
            db.CheckConstraint(
                '(fed_object_id IS NOT NULL AND fed_version_id IS NOT NULL AND component_id IS NOT NULL) OR (eln_import_id IS NOT NULL AND eln_object_id IS NOT NULL) OR (action_id IS NOT NULL AND data IS NOT NULL AND schema IS NOT NULL AND user_id IS NOT NULL AND utc_datetime IS NOT NULL)',
                name=table_name_prefix + '_previous_not_null_check'
            )
        )
        self._subversions_table = db.Table(
            table_name_prefix + '_subversions',
            self.metadata,
            db.Column('object_id', db.Integer, nullable=False),
            db.Column('version_id', db.Integer, nullable=False),
            db.Column('subversion_id', db.Integer, nullable=False),
            db.Column('action_id', db.Integer, nullable=True),
            db.Column('data', postgresql.JSONB, nullable=True),
            db.Column('schema', postgresql.JSONB, nullable=True),
            db.Column('user_id', db.Integer, nullable=True),
            db.Column('utc_datetime', db.TIMESTAMP(timezone=True), nullable=True),
            db.Column('utc_datetime_subversion', db.TIMESTAMP(timezone=True), nullable=False),
            db.PrimaryKeyConstraint('object_id', 'version_id', 'subversion_id')
        )
        if user_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['user_id'], [user_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['user_id'], [user_id_column]))
            self._subversions_table.append_constraint(db.ForeignKeyConstraint(['user_id'], [user_id_column]))
        if action_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['action_id'], [action_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['action_id'], [action_id_column]))
            self._subversions_table.append_constraint(db.ForeignKeyConstraint(['action_id'], [action_id_column]))
        elif action_schema_column is not None:
            raise ValueError("Using action_schema_column requires using the action_id_column as well!")
        if component_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['component_id'], [component_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['component_id'], [component_id_column]))
        if eln_import_id_column is not None:
            self._current_table.append_constraint(db.ForeignKeyConstraint(['eln_import_id'], [eln_import_id_column]))
            self._previous_table.append_constraint(db.ForeignKeyConstraint(['eln_import_id'], [eln_import_id_column]))
        self._action_id_column = action_id_column
        self._action_schema_column = action_schema_column
        self._component_id_column = component_id_column
        self._eln_import_id_column = eln_import_id_column
        self.object_id_column = self._current_table.c.object_id
        self.data_column = self._current_table.c.data
        self.bind = bind
        if self.bind is not None:
            self.metadata.create_all(self.bind)
        self._data_validator = data_validator
        self._schema_validator = schema_validator

    @_use_transaction
    def create_object(
            self,
            data: typing.Optional[typing.Dict[str, typing.Any]],
            schema: typing.Optional[typing.Dict[str, typing.Any]],
            user_id: typing.Optional[int],
            action_id: typing.Optional[int],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            fed_object_id: typing.Optional[int] = None,
            fed_version_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None,
            eln_import_id: typing.Optional[int] = None,
            eln_object_id: typing.Optional[str] = None,
            connection: typing.Optional[db.engine.Connection] = None,
            validate_data: bool = True,
            data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> Object:
        """
        Creates an object in the table for current objects. This object will always have version_id 0.

        :param data: a JSON serializable object containing the object data
        :param schema: a JSON schema describing data (may be None if action's schema is to be used)
        :param user_id: the ID of the user who created the object
        :param action_id: the ID of the action which was used to create the object
        :param utc_datetime: the datetime (in UTC) when the object was created (optional, defaults to now(timezone.utc))
        :param fed_object_id: the ID of this object on a federated component
        :param fed_version_id: the version ID of this object on a federated component
        :param eln_import_id: the ID of an .eln file import
        :param eln_object_id: the ID of this object in the .eln file
        :param component_id: the ID of the component that created this object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :param validate_data: whether the data must be validated
        :param data_validator_arguments: additional keyword arguments to the data validator
        :return: the newly created object as object_type
        :raise ValueError: if the schema and action are None for a local object
        """
        assert connection is not None  # ensured by decorator
        if utc_datetime is None and fed_object_id is None:
            utc_datetime = datetime.datetime.now(datetime.timezone.utc)
        if schema is not None or data is not None:
            if schema is None and action_id is not None and self._action_schema_column is not None and self._action_id_column is not None:
                action = connection.execute(
                    db
                    .select(
                        self._action_schema_column
                    )
                    .where(self._action_id_column == action_id)
                ).fetchone()
                if action is None:
                    if fed_object_id is None or fed_version_id is None or component_id is None:
                        raise ValueError(f'Action with id {action_id} not found')
                    else:
                        schema = None
                else:
                    schema = action[0]
            if not (schema is None and fed_object_id is not None and fed_version_id is not None):
                if schema is not None and self._schema_validator:
                    self._schema_validator(schema)
                if not (data is None and fed_object_id is not None and fed_version_id is not None):
                    if data is not None and schema is not None and self._data_validator and validate_data:
                        if data_validator_arguments is None:
                            data_validator_arguments = {}
                        if 'allow_disabled_languages' not in data_validator_arguments:
                            data_validator_arguments['allow_disabled_languages'] = False
                        self._data_validator(data, schema, **data_validator_arguments)
        version_id = 0
        object_id = typing.cast(int, connection.execute(
            self._current_table
            .insert()
            .values(
                version_id=version_id,
                action_id=action_id,
                data=data,
                name_cache=data.get('name', {}).get('text') if data else None,
                tags_cache=data.get('tags') if data else None,
                schema=schema,
                user_id=user_id,
                utc_datetime=utc_datetime,
                fed_object_id=fed_object_id,
                fed_version_id=fed_version_id,
                component_id=component_id,
                eln_import_id=eln_import_id,
                eln_object_id=eln_object_id,
            )
            .returning(
                self._current_table.c.object_id
            )
        ).scalar())
        obj = Object(
            object_id=object_id,
            version_id=version_id,
            action_id=action_id,
            data=data,
            schema=schema,
            user_id=user_id,
            utc_datetime=utc_datetime,
            fed_object_id=fed_object_id,
            fed_version_id=fed_version_id,
            component_id=component_id,
            eln_import_id=eln_import_id,
            eln_object_id=eln_object_id,
        )
        return obj

    @_use_transaction
    def update_object(
            self,
            object_id: int,
            data: typing.Optional[typing.Dict[str, typing.Any]],
            schema: typing.Optional[typing.Dict[str, typing.Any]],
            user_id: int,
            utc_datetime: typing.Optional[datetime.datetime] = None,
            connection: typing.Optional[db.engine.Connection] = None,
            validate_schema: bool = True,
            validate_data: bool = True,
            data_validator_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Optional[Object]:
        """
        Updates an existing object using the given data, user id and datetime.

        This step first copies the existing object in the table of previous object versions and then modifies the
        object's version id, incrementing it by one, and data, replacing it by the given data.

        :param object_id: the ID of the existing object
        :param data: a JSON serializable object containing the updated object data
        :param schema: a JSON schema describing data (optional, defaults to the current object schema)
        :param user_id: the ID of the user who updated the object
        :param utc_datetime: the datetime (in UTC) when the object was updated (optional, defaults to now(timezone.utc))
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :param validate_schema: whether the schema should be validated
        :param validate_data: whether the data should be validated
        :param data_validator_arguments: additional keyword arguments to the data validator
        :return: the updated object as object_type or None, if the object does not exist
        """
        assert connection is not None  # ensured by decorator
        if utc_datetime is None:
            utc_datetime = datetime.datetime.now(datetime.timezone.utc)
        if schema is not None or data is not None:
            if schema is None:
                schema_row = connection.execute(
                    db.select(self._current_table.c.schema).where(self._current_table.c.object_id == object_id)
                ).fetchone()
                if schema_row is None:
                    return None
                schema = schema_row[0]
            if validate_schema and self._schema_validator and schema is not None:
                self._schema_validator(schema)
            if validate_data and self._data_validator and schema is not None and data is not None:
                if data_validator_arguments is None:
                    data_validator_arguments = {}
                self._data_validator(data, schema, **data_validator_arguments)

        # Copy current version to previous versions
        if connection.execute(
            self._previous_table
            .insert()
            .from_select(
                [
                    'object_id',
                    'version_id',
                    'action_id',
                    'data',
                    'schema',
                    'user_id',
                    'utc_datetime',
                    'fed_object_id',
                    'fed_version_id',
                    'component_id',
                    'eln_import_id',
                    'eln_object_id',
                ],
                self._current_table
                .select()
                .with_only_columns(
                    self._current_table.c.object_id,
                    self._current_table.c.version_id,
                    self._current_table.c.action_id,
                    self._current_table.c.data,
                    self._current_table.c.schema,
                    self._current_table.c.user_id,
                    self._current_table.c.utc_datetime,
                    self._current_table.c.fed_object_id,
                    self._current_table.c.fed_version_id,
                    self._current_table.c.component_id,
                    self._current_table.c.eln_import_id,
                    self._current_table.c.eln_object_id,
                )
                .where(self._current_table.c.object_id == db.bindparam('oid'))
            ),
            [{'oid': object_id}]
        ).rowcount != 1:
            return None
        # Update current version to new version
        connection.execute(
            self._current_table
            .update()
            .where(self._current_table.c.object_id == db.bindparam('oid'))
            .values(
                version_id=self._current_table.c.version_id + 1,
                data=data,
                name_cache=data.get('name', {}).get('text') if data else None,
                tags_cache=data.get('tags') if data else None,
                schema=schema,
                user_id=user_id,
                utc_datetime=utc_datetime
            ),
            [{'oid': object_id}]
        )
        return self.get_current_object(object_id, connection=connection)

    @_use_transaction
    def restore_object_version(
            self,
            object_id: int,
            version_id: int,
            user_id: int,
            utc_datetime: typing.Optional[datetime.datetime] = None,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        object_version = self.get_object_version(object_id=object_id, version_id=version_id, connection=connection)
        if object_version is None:
            return None
        if object_version.schema is not None and self._schema_validator:
            self._schema_validator(object_version.schema)
        if object_version.schema is not None and object_version.data is not None and self._data_validator:
            # allow disabled languages, as they were enabled when the version was created
            self._data_validator(object_version.data, object_version.schema, allow_disabled_languages=True)
        return self.update_object(
            object_id=object_id,
            data=object_version.data,
            schema=object_version.schema,
            user_id=user_id,
            utc_datetime=utc_datetime,
            connection=connection,
            validate_schema=False,
            validate_data=False
        )

    @_use_transaction
    def insert_fed_object_version(
            self,
            fed_object_id: int,
            fed_version_id: int,
            component_id: int,
            action_id: typing.Optional[int],
            data: typing.Optional[typing.Dict[str, typing.Any]],
            schema: typing.Optional[typing.Dict[str, typing.Any]],
            user_id: typing.Optional[int],
            utc_datetime: typing.Optional[datetime.datetime],
            connection: typing.Optional[db.engine.Connection] = None,
            allow_disabled_languages: bool = False,
            get_missing_schema_from_action: bool = True
    ) -> typing.Optional[Object]:
        assert connection is not None  # ensured by decorator
        if schema is None and get_missing_schema_from_action and self._action_schema_column is not None and self._action_id_column is not None:
            action = connection.execute(
                db
                .select(
                    self._action_schema_column
                )
                .where(self._action_id_column == action_id)
            ).fetchone()
            if action is None:
                if fed_object_id is None or fed_version_id is None or component_id is None:
                    raise ValueError(f'Action with id {action_id} not found')
                else:
                    schema = None
            else:
                schema = action[0]
        validate_data = True
        if not (schema is None and fed_object_id is not None and fed_version_id is not None):
            if self._schema_validator:
                self._schema_validator(schema)
            if not (data is None and fed_object_id is not None and fed_version_id is not None):
                if self._data_validator:
                    self._data_validator(data, schema, allow_disabled_languages=allow_disabled_languages)
                    validate_data = False

        current = self.get_current_fed_object(component_id, fed_object_id)
        if current is None:
            object = self.create_object(data, schema, user_id, action_id, utc_datetime, fed_object_id, fed_version_id, component_id, connection=connection, validate_data=validate_data)
            return object
        elif current.fed_version_id is None or current.fed_version_id < fed_version_id:
            # Copy current version to previous versions
            if connection.execute(
                self._previous_table
                .insert()
                .from_select(
                    [
                        'object_id',
                        'version_id',
                        'action_id',
                        'data',
                        'schema',
                        'user_id',
                        'utc_datetime',
                        'fed_object_id',
                        'fed_version_id',
                        'component_id',
                        'eln_import_id',
                        'eln_object_id'
                    ],
                    self._current_table
                    .select()
                    .with_only_columns(
                        self._current_table.c.object_id,
                        self._current_table.c.version_id,
                        self._current_table.c.action_id,
                        self._current_table.c.data,
                        self._current_table.c.schema,
                        self._current_table.c.user_id,
                        self._current_table.c.utc_datetime,
                        self._current_table.c.fed_object_id,
                        self._current_table.c.fed_version_id,
                        self._current_table.c.component_id,
                        self._current_table.c.eln_import_id,
                        self._current_table.c.eln_object_id,
                    )
                    .where(self._current_table.c.object_id == db.bindparam('oid'))
                ),
                [{'oid': current.object_id}]
            ).rowcount != 1:
                return None
            # Update current version to new version
            connection.execute(
                self._current_table
                .update()
                .where(self._current_table.c.object_id == db.bindparam('oid'))
                .values(
                    object_id=current.object_id,
                    version_id=self._current_table.c.version_id + 1,
                    data=data,
                    name_cache=data.get('name', {}).get('text') if data else None,
                    tags_cache=data.get('tags') if data else None,
                    schema=schema,
                    action_id=action_id,
                    user_id=user_id,
                    utc_datetime=utc_datetime,
                    fed_object_id=fed_object_id,
                    fed_version_id=fed_version_id,
                    component_id=component_id
                ),
                [{'oid': current.object_id}]
            )
        else:
            max_version_id = connection.execute(
                db.select(
                    db.sql.func.max(self._previous_table.c.version_id)  # pylint: disable=not-callable
                ).where(
                    self._previous_table.c.object_id == current.object_id
                )
            ).scalar()
            if max_version_id is None:
                max_version_id = -1
            if connection.execute(
                self._previous_table
                .insert()
                .values(
                    object_id=current.object_id,
                    version_id=max(max_version_id, current.version_id) + 1,
                    action_id=action_id,
                    data=data,
                    schema=schema,
                    user_id=user_id,
                    utc_datetime=utc_datetime,
                    fed_object_id=fed_object_id,
                    fed_version_id=fed_version_id,
                    component_id=component_id
                ),
                [{'oid': current.object_id}]
            ).rowcount != 1:
                return None
        return self.get_fed_object_version(component_id, fed_object_id, fed_version_id, connection=connection)

    @_use_transaction
    def update_object_version(
            self,
            object_id: int,
            version_id: int,
            action_id: typing.Optional[int],
            schema: typing.Optional[typing.Dict[str, typing.Any]],
            data: typing.Optional[typing.Dict[str, typing.Any]],
            user_id: typing.Optional[int],
            utc_datetime: typing.Optional[datetime.datetime],
            utc_datetime_subversion: typing.Optional[datetime.datetime] = None,
            connection: typing.Optional[db.engine.Connection] = None,
            allow_disabled_languages: bool = False,
            get_missing_schema_from_action: bool = True
    ) -> typing.Optional[Object]:
        """
        Updates an existing object version using the given data, user id and datetime.

        This step first copies the existing object version in the table of object subversions and then modifies the
        data, replacing it by the given data.

        :param object_id: the ID of the existing object
        :param version_id: the version ID of the existing object
        :param action_id: the ID of the existing action
        :param data: a JSON serializable object containing the updated object data
        :param schema: a JSON schema describing data (optional, defaults to the current object schema)
        :param user_id: the ID of the user who updated the object
        :param utc_datetime: the datetime (in UTC) when the object was updated
        :param utc_datetime_subversion: the datetime (in UTC) when the object update has been applied (optional, defaults to now(timezone.utc))
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :param allow_disabled_languages: whether using disabled languages should be allowed in this update
        :param get_missing_schema_from_action: whether to use an action schema (if available) when None is passed for schema
        :return: the updated object as object_type or None, if the object does not exist
        """
        assert connection is not None  # ensured by decorator
        if utc_datetime_subversion is None:
            utc_datetime_subversion = datetime.datetime.now(datetime.timezone.utc)
        if schema is None and get_missing_schema_from_action and self._action_schema_column is not None and self._action_id_column is not None:
            action = connection.execute(
                db
                .select(
                    self._action_schema_column
                )
                .where(self._action_id_column == action_id)
            ).fetchone()
            if action is None:
                schema = None
            else:
                schema = action[0]
        if schema is not None:
            if self._schema_validator:
                self._schema_validator(schema)
            if data is not None:
                if self._data_validator:
                    self._data_validator(data, schema, allow_disabled_languages=allow_disabled_languages)

        assert connection is not None  # ensured by decorator
        previous_objects = connection.execute(
            db
            .select(
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime,
                self._previous_table.c.fed_object_id,
                self._previous_table.c.fed_version_id,
                self._previous_table.c.component_id,
                self._previous_table.c.eln_import_id,
                self._previous_table.c.eln_object_id,
            )
            .where(db.and_(
                self._previous_table.c.object_id == object_id,
                self._previous_table.c.version_id == version_id
            ))
        ).fetchall()
        if previous_objects:
            selected_table = self._previous_table
            object_data = Object(*previous_objects[0])
        else:
            current_object = self.get_current_object(object_id, connection=connection)
            if current_object is not None and current_object.version_id == version_id:
                selected_table = self._current_table
                object_data = current_object
            else:
                return None
        if object_data.fed_object_id is None or object_data.fed_version_id is None or object_data.component_id is None:
            return None

        previous_subversion_id = connection.execute(
            db.select(
                db.sql.func.max(self._subversions_table.c.subversion_id)  # pylint: disable=not-callable
            ).where(
                db.and_(
                    self._subversions_table.c.object_id == object_id,
                    self._subversions_table.c.version_id == version_id
                )
            )
        ).scalar()
        if previous_subversion_id is None:
            subversion_id = 0
        else:
            subversion_id = previous_subversion_id + 1
        if connection.execute(
            # move deprecated version to subversions table
            self._subversions_table
                .insert()
                .values(
                    object_id=object_data.object_id,
                    version_id=object_data.version_id,
                    subversion_id=subversion_id,
                    action_id=object_data.action_id,
                    data=object_data.data,
                    schema=object_data.schema,
                    user_id=object_data.user_id,
                    utc_datetime=object_data.utc_datetime,
                    utc_datetime_subversion=utc_datetime_subversion
                )
        ).rowcount != 1:
            return None

        # Update current version to new version
        if selected_table == self._current_table:
            cache_values = {
                'name_cache': data.get('name', {}).get('text') if data else None,
                'tags_cache': data.get('tags') if data else None,
            }
        else:
            cache_values = {}
        connection.execute(
            selected_table
            .update()
            .where(selected_table.c.object_id == db.bindparam('oid'), selected_table.c.version_id == db.bindparam('vid'))
            .values(
                action_id=action_id,
                data=data,
                schema=schema,
                user_id=user_id,
                utc_datetime=utc_datetime,
                **cache_values
            ),
            [{'oid': object_id, 'vid': version_id}]
        )
        return self.get_object_version(object_id, version_id, connection=connection)

    @_use_transaction
    def is_existing_object(
            self,
            object_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> bool:
        """
        Return whether an object with the given ID exists.

        :param object_id: the ID of a possibly existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: whether the object exists
        """
        assert connection is not None  # ensured by decorator
        return connection.execute(
            db.select(self._current_table.c.object_id).where(self._current_table.c.object_id == object_id)
        ).fetchone() is not None

    @_use_transaction
    def is_existing_object_version(
            self,
            object_id: int,
            version_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> bool:
        """
        Return whether an object version with the given ID exists.

        :param object_id: the ID of a possibly existing object
        :param version_id: the version ID to check
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: whether the object version exists
        """
        assert connection is not None  # ensured by decorator
        current_version_id = connection.execute(
            db.select(self._current_table.c.version_id).where(
                self._current_table.c.object_id == object_id,
                self._current_table.c.version_id >= version_id
            )
        ).fetchone()
        if current_version_id is None:
            return False
        if current_version_id[0] == version_id:
            return True
        return connection.execute(
            db.select(self._previous_table.c.version_id).where(
                self._previous_table.c.object_id == object_id,
                self._previous_table.c.version_id == version_id
            )
        ).fetchone() is not None

    @_use_transaction
    def get_current_object(
            self,
            object_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        """
        Queries and returns an object by its ID.

        :param object_id: the ID of the existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the object as object_type or None if the object does not exist
        """
        assert connection is not None  # ensured by decorator
        current_object = connection.execute(
            db
            .select(
                self._current_table.c.object_id,
                self._current_table.c.version_id,
                self._current_table.c.action_id,
                self._current_table.c.data,
                self._current_table.c.schema,
                self._current_table.c.user_id,
                self._current_table.c.utc_datetime,
                self._current_table.c.fed_object_id,
                self._current_table.c.fed_version_id,
                self._current_table.c.component_id,
                self._current_table.c.eln_import_id,
                self._current_table.c.eln_object_id,
            )
            .where(self._current_table.c.object_id == object_id)
        ).fetchone()
        if current_object is None:
            return None
        return Object(*current_object)

    @_use_transaction
    def get_previous_subversion(
            self,
            object_id: int,
            version_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        assert connection is not None  # ensured by decorator
        previous_object_subversion = connection.execute(
            db.select(
                self._subversions_table.c.object_id,
                self._subversions_table.c.version_id,
                self._subversions_table.c.action_id,
                self._subversions_table.c.data,
                self._subversions_table.c.schema,
                self._subversions_table.c.user_id,
                self._subversions_table.c.utc_datetime,
                db.null(),  # fed_object_id
                db.null(),  # fed_version_id
                db.null(),  # component_id
                db.null(),  # eln_import_id
                db.null(),  # eln_object_id
            )
            .where(
                db.and_(
                    self._subversions_table.c.object_id == object_id,
                    self._subversions_table.c.version_id == version_id
                )
            )
            .order_by(db.desc(self._subversions_table.c.subversion_id))
        ).first()
        if previous_object_subversion is None:
            return None
        return Object(*previous_object_subversion)

    @_use_transaction
    def get_current_fed_object(
            self,
            component_id: int,
            fed_object_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        """
        """
        assert connection is not None  # ensured by decorator
        current_object = connection.execute(
            db
            .select(
                self._current_table.c.object_id,
                self._current_table.c.version_id,
                self._current_table.c.action_id,
                self._current_table.c.data,
                self._current_table.c.schema,
                self._current_table.c.user_id,
                self._current_table.c.utc_datetime,
                self._current_table.c.fed_object_id,
                self._current_table.c.fed_version_id,
                self._current_table.c.component_id,
                self._current_table.c.eln_import_id,
                self._current_table.c.eln_object_id,
            )
            .where(db.and_(
                self._current_table.c.component_id == component_id,
                self._current_table.c.fed_object_id == fed_object_id
            ))
        ).fetchone()
        if current_object is None:
            return None
        return Object(*current_object)

    @_use_transaction
    def get_fed_object_version(
            self,
            component_id: int,
            fed_object_id: int,
            fed_version_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        """
        """
        assert connection is not None  # ensured by decorator
        previous_objects = connection.execute(
            db
            .select(
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime,
                self._previous_table.c.fed_object_id,
                self._previous_table.c.fed_version_id,
                self._previous_table.c.component_id,
                self._previous_table.c.eln_import_id,
                self._previous_table.c.eln_object_id,
            )
            .where(db.and_(db.and_(
                self._previous_table.c.component_id == component_id,
                self._previous_table.c.fed_object_id == fed_object_id),
                self._previous_table.c.fed_version_id == fed_version_id
            ))
        ).fetchall()
        if previous_objects:
            return Object(*previous_objects[0])
        current_object = self.get_current_fed_object(component_id, fed_object_id, connection=connection)
        if current_object is not None and current_object.fed_version_id == fed_version_id:
            return current_object
        return None

    @_use_transaction
    def get_current_objects(
            self,
            filter_func: typing.Callable[[typing.Any], typing.Any] = lambda data: True,
            action_table: typing.Any = None,
            action_filter: typing.Optional[db.sql.ColumnElement[bool]] = None,
            connection: typing.Optional[db.engine.Connection] = None,
            table: typing.Any = None,
            parameters: typing.Optional[typing.Dict[str, typing.Any]] = None,
            sorting_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], typing.Any]] = None,
            limit: typing.Optional[int] = None,
            offset: typing.Optional[int] = None,
            num_objects_found: typing.Optional[typing.List[int]] = None
    ) -> typing.List[Object]:
        """
        Queries and returns all objects matching a given filter.

        :param filter_func: a lambda that may return an SQLAlchemy filter when given a table
        :param action_table: a SQLAlchemy table object containing the actions to filter by (see action_filter)
        :param action_filter: a SQLAlchemy comparator, used to query only objects created by specific actions
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :param table: a custom SQLAlchemy table-like object to use as base for the query (optional)
        :param parameters: query parameters for the custom select statement (optional)
        :param sorting_func: a sorting function to use (see logic.object_sorting)
        :param limit: limits the number of returned objects, if set
        :param offset: an offset to apply to the query
        :param num_objects_found: a list used to return the number of objects found in, using the 0-th element
        :return: a list of objects as object_type
        """
        assert connection is not None  # ensured by decorator

        if parameters is None:
            parameters = {}

        if table is None:
            table = self._current_table

        select_statement = db.select(
            table.c.object_id,
            table.c.version_id,
            table.c.action_id,
            table.c.data,
            table.c.schema,
            table.c.user_id,
            table.c.utc_datetime,
            table.c.fed_object_id,
            table.c.fed_version_id,
            table.c.component_id,
            table.c.eln_import_id,
            table.c.eln_object_id,
            db.sql.expression.text('COUNT(*) OVER()')
        )

        selectable = table

        if sorting_func is not None and getattr(sorting_func, 'require_original_columns', False):
            selectable = selectable.outerjoin(
                self._previous_table,
                db.and_(table.c.object_id == self._previous_table.c.object_id, self._previous_table.c.version_id == 0),
                full=False
            )

        if action_table is not None and action_filter is not None:
            assert self._action_id_column is not None
            assert action_table is not None
            assert action_filter is not None
            selectable = selectable.join(
                action_table,
                db.and_(table.c.action_id == self._action_id_column, action_filter)
            )

        select_statement = select_statement.select_from(selectable)

        if sorting_func is None:
            def default_sorting_func(current_columns: typing.Any, original_columns: typing.Any) -> typing.Any:
                return db.sql.desc(current_columns.object_id)

            sorting_func = default_sorting_func

        # use data_full column for filtering, as data might only contain the name property
        filter_data_column = table.c.data_full if hasattr(table.c, 'data_full') else table.c.data
        # set object_id_column to allow access to table.c.object_id in filter_func (e.g. for file search)
        filter_data_column.object_id_column = table.c.object_id
        select_statement = select_statement.where(filter_func(filter_data_column))
        select_statement = select_statement.order_by(sorting_func(table.c, self._previous_table.c))

        if limit is not None:
            select_statement = select_statement.limit(limit)

        if offset is not None:
            select_statement = select_statement.offset(offset)

        objects = connection.execute(
            select_statement,
            parameters
        ).fetchall()
        if num_objects_found is not None:
            num_objects_found.clear()
            if objects:
                num_objects_found.append(objects[0][-1])
            else:
                num_objects_found.append(0)
        return [Object(*obj[:-1]) for obj in objects]

    @_use_transaction
    def get_object_versions(
            self,
            object_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.List[Object]:
        """
        Queries and returns all versions of an object with a given ID, sorted ascendingly by the version ID, from first
        version to current version.

        :param object_id: the ID of the existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: a list of objects as object_type
        """
        assert connection is not None  # ensured by decorator
        current_object = self.get_current_object(object_id, connection=connection)
        if current_object is None:
            return []
        previous_objects = connection.execute(
            db
            .select(
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime,
                self._previous_table.c.fed_object_id,
                self._previous_table.c.fed_version_id,
                self._previous_table.c.component_id,
                self._previous_table.c.eln_import_id,
                self._previous_table.c.eln_object_id,
            )
            .where(self._previous_table.c.object_id == object_id)
            # .order_by(db.asc(self._previous_table.c.version_id))
            .order_by(db.asc(self._previous_table.c.utc_datetime))
        ).fetchall()
        objects = []
        for obj in previous_objects:
            objects.append(Object(*obj))
        objects.append(current_object)
        return objects

    @_use_transaction
    def get_object_version(
            self,
            object_id: int,
            version_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[Object]:
        """
        Queries and returns an individual version of an object with a given object and version ID.

        :param object_id: the ID of the existing object
        :param version_id: the ID of the object's existing version
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the object as object_type or None if the object or this version of it does not exist
        """
        assert connection is not None  # ensured by decorator
        previous_objects = connection.execute(
            db
            .select(
                self._previous_table.c.object_id,
                self._previous_table.c.version_id,
                self._previous_table.c.action_id,
                self._previous_table.c.data,
                self._previous_table.c.schema,
                self._previous_table.c.user_id,
                self._previous_table.c.utc_datetime,
                self._previous_table.c.fed_object_id,
                self._previous_table.c.fed_version_id,
                self._previous_table.c.component_id,
                self._previous_table.c.eln_import_id,
                self._previous_table.c.eln_object_id,
            )
            .where(db.and_(
                self._previous_table.c.object_id == object_id,
                self._previous_table.c.version_id == version_id
            ))
        ).fetchall()
        if previous_objects:
            return Object(*previous_objects[0])
        current_object = self.get_current_object(object_id, connection=connection)
        if current_object is not None and current_object.version_id == version_id:
            return current_object
        return None

    @_use_transaction
    def get_current_object_version_id(
            self,
            object_id: int,
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Optional[int]:
        """
        Get the ID of the current version of an object.

        :param object_id: the ID of an existing object
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the object's current version ID, or None
        """
        assert connection is not None  # ensured by decorator
        current_object_info = connection.execute(
            db
            .select(
                self._current_table.c.version_id,
            )
            .where(
                self._current_table.c.object_id == object_id
            )
        ).fetchone()
        if current_object_info is None:
            return None
        return typing.cast(int, current_object_info[0])

    @_use_transaction
    def get_action_ids_for_object_ids(
            self,
            object_ids: typing.Sequence[int],
            connection: typing.Optional[db.engine.Connection] = None
    ) -> typing.Dict[int, typing.Optional[int]]:
        """
        Get the action IDs for a list of object IDs.

        :param object_ids: the IDs of existing objects
        :param connection: the SQLAlchemy connection (optional, defaults to a new connection using self.bind)
        :return: the objects' action IDs
        """
        assert connection is not None  # ensured by decorator

        object_ids_and_action_ids = connection.execute(
            db
            .select(
                self._current_table.c.object_id, self._current_table.c.action_id
            )
            .where(
                self._current_table.c.object_id.in_(object_ids)
            )
        ).fetchall()
        result = {
            object_id: action_id
            for object_id, action_id in object_ids_and_action_ids
        }
        for object_id in object_ids:
            if object_id not in result:
                result[object_id] = None
        return result

    @property
    def current_table(self) -> db.Table:
        return self._current_table
